#!/usr/bin/perl

# Snortsnarf, a utility to convert snort log files to HTML pages
# Authors: Stuart Staniford, Silicon Defense (stuart@SiliconDefense.com)
#          James Hoagland, Silicon Defense (hoagland@SiliconDefense.com)
# copyright (c) 2000 by Silicon Defense (http://www.silicondefense.com/)

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.  

# Snortsnarf description:
#
# Code to parse files of snort alerts and portscan logs, and produce
# HTML output intended to allow an intrusion detection analyst to
# engage in diagnostic inspection and tracking down problems.  
# The model is that one is using a cron job or similar to
# produce a daily/hourly/whatever file of snort alerts.	 This script
# can be run on each such file to produce a convenient HTML breakout
# of all the alerts.
#
# The idea is that the analyst can click around the alerts instead
# of wearily grepping and awking through them.

# Please send complaints, kudos, and especially improvements and bugfixes to
# hoagland@SiliconDefense.com. This is a quick hack with features added and
# may only be worth what you paid for it.  It is still under active
# development and may change at any time.

# this file (snortsnarf.pl) is part of Snortsnarf v090700.1

##############################################################################

# Usage:

# snortsnarf.pl <options> <file1 file2 ...>

# Any filenames with the string 'scan' in the name are assumed to be
# portscan files rather than alert files.  Similarly, filenames containing
# the string 'syslog' are assumed to be syslog files.

# The script will produce a directory snfout.file1 (by default) full of
# a large number of html files.	 These are placed under the current
# directory.  It also produces a file index.html in that directory.	 
# This is a good place to start with a browser.

# Options include:

# -d directory			directory is the path to the directory the HTML pages will
#							be generated in, overriding the default.

# -dns					This will cause the script to lookup the DNS name 
#								of each IP it writes a page for.  On a large alert 
#								file, this will take a very long time and will hammer 
#								your DNS server.

# -ldir URL				URL is the URL of a base directory in which the 
#								log files usually found in /var/log/snort are living,
#								With this option, snortsnarf will generate lots of 
#								links into those files based at that URL.  Eg: with
#								-ldir "http://host.name.here/logs" we get links like
#								http://host.name.here/logs/10.0.0.1/UDP-137-137
#								Note that the logs themselves aren't parsed.

# -homenet network		network is the network IP address or CIDR network notation
#								for your home network, as told to snort.  CIDR takes
#								that standard form of address/masksize (sizes 0-32
#								supported).  In the other form, 1-3 zeros at the end
#								of the address are assumed to be what varies within
#								your network.	 If this argument is not provided, the
#								home network is assumed to be 0.0.0.0 (no home).  This
#								is currently used only with the -ldir option.

# -color=<option>		option is 'yes' or 'no' or 'rotate'.  The style of coloring
#								is given by the option if any.  'rotate' is the only
#								style currently available.  This changes the background
#								color between an alert/portscan report and the next one
#								iff the next one has a different message (signature)
#								source host, or destination host.  If 'yes' or this not
#								given on the command line, the the default ('rotate' in
#								this version) is used. 

# -split=<threshold>	To speed up display of pages containing lots of alerts,
#							<threshold> is a maximum number of alerts to display on a
#							page unless explicitly requested.  Instead the alerts are
#							broken into several pages.  Default is 100.  Can be set to
#							0 to never split the list alerts being displayed.

# -onewindow			If this option is given, certain URs will not be targeted to
#							other browser windows as is the default.

# -db path				path is the full path to the annotation database (an XML file),
#							from the point of view of the server hosting the CGI scripts
#							that access it, or the empty string to not use an annotation
#							database.  The default is to not use it.

# -cgidir URL 			URL is the location of the cgi-bin directory that CGI scripts
#							than go along with the program are stored.  This may be
#							relative to the pages that are generated (so that when a
#							link appears on a page with URL as a prefix, it will be
#							valid).  "/cgi-bin" is the default.

# -sisr configfile		Generate links with SnortSnarf Incident Storage and Reporting 
#							(SISR).  The argument is the full path to the SISR
#							configuration file to use.  This file is not parsed by
#							snortsnarf.pl.

# -nmapurl URL			URL is the URL of a base URL directory in which the html
#							output of running nmap2html on nmap output is living. 
#							With this option, snortsnarf will generate links to that
#							output for IP addresses.  If the path to that directory
#							is given with the -nmapdir option, then only those pages
#							that actually exist (at the time snortsnarf is run) are
#							linked to.

# -nmapdir dir			dir is the directory on the file system in which nmap2html
#							output is stored.  When the -nmapurl option is given,
#							this is used to verify that a nmaplog.pl page actually
#							exists before linking to it. 

# -rulesfile file		file is the base rule file (e.g., snort.lib) that snort was
#							run with.  If this option is given, the rulesin that
#							file (and included files) generating a signature  are
#							included in the page for that signatures alerts.  Note
#							that the processing of the rule files are pretty rough.
#							E.g., variables are ignored. 

# -rulesdir dir			dir is a directry to use as the base path for rules files
#							rather than the one given or the paths listed in include
#							directives.  (Useful if the files have been relocated.)


# Snortsnarf is believed to work with Perl 5 on the following OS's
# OpenBSD 2.6
# RedHat Linux 6.1 and 6.2
# Windows NT 4.0 (NTFS only) (tweak the $os parameter below.  Wildcards won't).

# It probably works with most other Unixen/Linuxen, and is certain not to 
# work on the FAT filesystem.

# Believed to work with alert files generated with these snort command lines:
# snort -D -c oursnort.lib -o -d 
# snort -i eth0 -c oursnort.lib -o -e  -d

# This script is pretty memory intensive while it is running.  A previous
# version of the script was known to handle 5MB of alerts on a machine with
# 64MB of memory (no virtual memory).  The script should be studied to figure
# out why it uses more memory than expected.

# HTML looks marginally ok with Netscape 4.7 and Explorer 4.72.	 
# Other browsers not tested.

# Tested on alert files up to about 1MB.  
# HTML takes up 5 times the space of the original alert file.  
# .tar.gz of the dir is about half the size of the alert file.
# YMMV!

##############################################################################

# Credits, etc

# Initial alert parsing code borrowed from Joey McAlerney, Silicon 
# Defense.
# A couple of ideas were stolen from Snort2html by Dan Swan.
# Thanks to SANS GIAC and Donald McLachlan for a paper with ideas 
# on how to look up IP addresses.

# Huge thanks to DARPA for supporting us for part of the time developing
# this code.  Major thanks to Paul Arabelo for being our test customer
# while we learned to do operational intrusion detection instead of being
# only a researcher.

# Shouts to Marty Roesch and Patrick Mullen and the rest of the snort
# community for developing snort and the snort portscan preprocessor.  

# Kudos to Steve Northcutt who has taught us all what an intrusion
# detection analyst should be.

# Version control info: $Id: snortsnarf.pl,v 1.25 2000/09/08 01:27:43 jim Exp $

$script = "<a href=\"http://www.silicondefense.com/snortsnarf/\">".
						"Snortsnarf</a>";
$version = "v090700.1";
$author_email = "hoagland\@SiliconDefense.com";
$author = "<a href=\"mailto:hoagland\@SiliconDefense.com\">Jim Hoagland</a> and <a href=\"mailto:stuart\@SiliconDefense.com\">Stuart Staniford</a>";

use Socket;
#use Sys::Hostname;
use Cwd;
require "snort_alert_parse.pl";
require "web_utils.pl";

##############################################################################

# portability stuff - toggle for Unix/Windows.

$os = 'unix';  # Either 'windows' or 'unix'
if($os eq 'windows')
{
 $dirsep = "\\";		
 #$root = "d:\\";		
 $root = "e:\\";
}
elsif($os eq 'unix')
{
 $dirsep = "\/";		# Unix
 $root = "\/";			# Unix
}

$html = 'html';			# usually html or htm

# Various global variables

$alert_file = $root."var".$dirsep."log".$dirsep."snort.alert"; # default input file
$index_page = "index.$html";
$main_page = "index.$html";

%monthnum = ('Jan' => 1,
			 'Feb' => 2,
			 'Mar' => 3,
			 'Apr' => 4,
			 'May' => 5,
			 'Jun' => 6,
			 'Jul' => 7,
			 'Aug' => 8,
			 'Sep' => 9,
			 'Oct' => 10,
			 'Nov' => 11,
			 'Dec' => 12);

# maps the text for ICMP messages found in an alert message to its component
# in a file name for an snort log of the connection
%ICMP_text_to_filename= ( 
		'ECHO REPLY' => 'ICMP_ECHO_REPLY',
		'DESTINATION UNREACHABLE: NET UNREACHABLE' => 'ICMP_NET_UNRCH', 
		'DESTINATION UNREACHABLE: HOST UNREACHABLE' => 'ICMP_HST_UNRCH', 
		'DESTINATION UNREACHABLE: PROTOCOL UNREACHABLE' => 'ICMP_PROTO_UNRCH', 
		'DESTINATION UNREACHABLE: PORT UNREACHABLE' => 'ICMP_PORT_UNRCH', 
		'DESTINATION UNREACHABLE: FRAGMENTATION NEEDED' => 'ICMP_UNRCH_FRAG_NEEDED', 
		'DESTINATION UNREACHABLE: SOURCE ROUTE FAILED' => 'ICMP_UNRCH_SOURCE_ROUTE_FAILED', 
		'DESTINATION UNREACHABLE: NET UNKNOWN' => 'ICMP_UNRCH_NETWORK_UNKNOWN', 
		'DESTINATION UNREACHABLE: HOST UNKNOWN' => 'ICMP_UNRCH_HOST_UNKNOWN', 
		'DESTINATION UNREACHABLE: HOST ISOLATED' => 'ICMP_UNRCH_HOST_ISOLATED', 
		'DESTINATION UNREACHABLE: NET ANO' => 'ICMP_UNRCH_NET_ANO',
		'DESTINATION UNREACHABLE: HOST ANO' => 'ICMP_UNRCH_HOST_ANO',
		'DESTINATION UNREACHABLE: NET UNREACHABLE TOS' => 'ICMP_UNRCH_NET_UNR_TOS', 
		'DESTINATION UNREACHABLE: HOST UNREACHABLE TOS' => 'ICMP_UNRCH_HOST_UNR_TOS', 
		'DESTINATION UNREACHABLE: PACKET FILTERED' => 'ICMP_UNRCH_PACKET_FILT', 
		'DESTINATION UNREACHABLE: PREC VIOLATION' => 'ICMP_UNRCH_PREC_VIOL', 
		'DESTINATION UNREACHABLE: PREC CUTOFF' => 'ICMP_UNRCH_PREC_CUTOFF', 
		'DESTINATION UNREACHABLE: UNKNOWN' => 'ICMP_UNKNOWN', 
		'SOURCE QUENCH' => 'ICMP_SRC_QUENCH', 
		'REDIRECT' => 'ICMP_REDIRECT', 
		'ECHO' => 'ICMP_ECHO', 
		'TTL EXCEEDED' => 'ICMP_TTL_EXCEED', 
		'PARAMATER PROBLEM' => 'ICMP_PARAM_PROB', 
		'TIMESTAMP REQUEST'=> 'ICMP_TIMESTAMP', 
		'TIMESTAMP REPLY' => 'ICMP_TIMESTAMP_RPL', 
		'INFO REQUEST' => 'ICMP_INFO_REQ', 
		'INFO REPLY' => 'ICMP_INFO_RPL', 
		'ADDRESS REQUEST'=> 'ICMP_ADDR', 
		'ADDRESS REPLY' => 'ICMP_ADDR_RPL', 
		'UNKNOWN' => 'ICMP_UNKNOWN' 
);

# the colors to cycle through for displayed alerts
@color= ('#E7DEBD','#E0CDD0','#D5E2CE');

$normbgcol= '#E7DEBD'; # normal background color
$anombgcol= '#B0E0F0'; # background color for anomaly pages

##############################################################################

# Main program

&process_options();
&initialize();

# gather alerts from input files
foreach $current_file (@ARGV)
{
	my $file_type;
	if($current_file =~ /scan/)
	 { $file_type = 'scan';}
	elsif($current_file =~ /syslog/)
	 { $file_type = 'syslog';}
	elsif($current_file =~ /messages/)
	 { $file_type = 'syslog';}
	else
	 { $file_type = 'alert';}

	my $fh= &initialize_per_file($current_file);

	while($alert = &get_alert($fh,$file_type))
	 {
	 	$alert->{'anomscore'}= &is_anom_rept($alert);
	 	if (defined($alert->{'anomscore'})) {
	   		&process_anom_alert($alert);
	 	} else {
	   		&process_alert($alert);
	   	}
	 }
	&close_out_file($fh);
}

# generate pages
$bgcol= $normbgcol;
&construct_sig_indexes();
&output_main_sig_page();
&output_per_sig();
&output_per_source();
&output_per_dest();

$bgcol= $anombgcol;
&output_anom_sect();


##############################################################################

# process the command line options and leave @ARGV with just the input files
# at the end
sub process_options
{
  my $arg;

  # default values
  $dns_option= undef;
  $log_base= '';
  $homenetaddr= undef;
  $color_opt= 'rotate';
  $cgi_dir= '/cgi-bin';
  $db_file= '';
  $split_threshold= 100;
  $nmap_dir= undef;
  $nmap_url= undef;
  $sisr_config= undef;
  $rules_file= undef;
  $rules_dir= undef;
  $notarget_option= 0;
  # go through arguments
  while($ARGV[0] =~ /^\-/)
   {
	$arg = shift @ARGV;
	if($arg eq '-dns')
	 {
		  $dns_option = 1;
	 }
	elsif($arg eq '-ldir')
	 {
		  $log_base = shift @ARGV;
		  $log_base.='/' unless $log_base =~ /\/$/;
	 }
	elsif($arg eq '-homenet')
	 {
	 	  my($addr,$bits)= split('/',(shift @ARGV));
	 	  $homenetaddr= &bytenums2bits(split('\.',$addr));
	 	  unless (defined($bits)) {
	 	  	if ($addr =~ /^0\.0\.0\.0$/) {
	 	  		$bits= 0;
	 	  	} elsif ($addr =~ /\.0\.0\.0$/) {
	 	  		$bits= 8;
	 	  	} elsif ($addr =~ /\.0\.0$/) {
	 	  		$bits= 16;
	 	  	} elsif ($addr =~ /\.0$/) {
	 	  		$bits= 24;
	 	  	} else {
	 	  		$bits= 32;
	 	  	}
	 	  }
	 	  my (@bytembits)= ();
	 	  foreach (1..4) {
	 	  	if ($bits <= 0) {
	 	  		push(@bytembits,0);
	 	  	} else {
	 	  		$bits-= 8;
	 	  		if ($bits >= 0) {
	 	  			push(@bytembits,255);
	 	  		} else {
	 	  			push(@bytembits,(0x80,0xC0,0xE0,0xF0,0xF8,0xFC,0xFE)[$bits+7]);
	 	  		}
	 	  	}
	 	  }
	 	  $homenetmask= &bytenums2bits(@bytembits);
		  $homenetaddr &= $homenetmask;
	 }
    elsif($arg =~ s/^-color//)
     {    
		if ($arg =~ /=(.*)/) {
			$color_opt = ($1 eq 'yes')?'rotate':$1;
		} else {
			$color_opt = 'rotate';
		}
	 }
    elsif($arg =~ s/^-split=//)
     {    
		$split_threshold = $arg;
	 }
	elsif($arg eq '-d')
	 {
		  $output_dir = shift @ARGV;
	 }
	elsif($arg eq '-cgidir')
	 {
		  $cgi_dir = shift @ARGV;
	 }
	elsif($arg eq '-db')
	 {
		  $db_file = shift @ARGV;
	 }
	elsif($arg eq '-nmapdir')
	 {
		  $nmap_dir = shift @ARGV;
	 }
	elsif($arg eq '-nmapurl')
	 {
		  $nmap_url = shift @ARGV;
		  $nmap_url.='/' unless $nmap_url =~ /\/$/;
	 }
	elsif($arg eq '-sisr')
	 {
		  $sisr_config = shift @ARGV;
	 }
	elsif($arg eq '-rulesfile')
	 {
		  $rules_file = shift @ARGV;
	 }
	elsif($arg eq '-rulesdir')
	 {
		  $rules_dir = shift @ARGV;
	 }
	elsif($arg eq '-onewindow')
	 {
		  $notarget_option = 1;
	 }
	else
	 {
	  print "Unknown option $arg\n";
	 }
   }
}

##############################################################################

sub initialize
{
	# Setup to use default file if no args
	if(@ARGV == 0)
	 {
		@ARGV = ($alert_file);
	 }
  	$count = 0;
  
  	$fhnonce='fh00';
  
 	$cwd= getcwd();
  
 	# Setup an output directory named after the first file (unless defined on command line)
  
	my $last = rindex($ARGV[0],$dirsep);
	my $substr = substr($ARGV[0],$last+1);
	$file_name[$count] = $substr;
	$output_dir = "snfout.$file_name[$count]" unless defined($output_dir);
	$anomoutdir= $output_dir.$dirsep.'anomrep';
  
	if(-e $output_dir)
	 {
		die("$output_dir already exists and is not a directory\n")
			unless -d $output_dir;
	  	if (opendir(D,$output_dir)) {
	  		while ($file=readdir(D)) {
				next unless $file =~ /^((src|dest)\d+\.\d+\.\d+\.\d+.*|(sig\d+)|index)\.$html/;
				# looks like a file we created, so delete it
				$fullpath= "$output_dir$dirsep$file";
				unlink($fullpath) || warn "could not clear $fullpath in output directory";
			}
			closedir(D);
	  	} else {
			warn "could not open output directory $output_dir to clear it";
	  	}
   }
  else
   {
	mkdir($output_dir,0755);
   }
}

##############################################################################

sub initialize_per_file
{
  my($file) = @_;

  $alert_file = $file;
  $file_fullpath[$count]= $file;
  $file_fullpath[$count]= "$cwd/$file" unless $file =~ /^\//;
  my $fh='alertfh00';
  open($fh,"$alert_file") || die("Couldn't open $file_type input file ".
														"$alert_file\n");
  my $last = rindex($alert_file,$dirsep);
  my $substr = substr($alert_file,$last+1);
  $file_name[$count] = $substr;
  return $fh;
}

##############################################################################

sub close_out_file
{
  close($_[0]);
  $count++;
}

##############################################################################

# return the next alert from the given file handle (expected to be in the
# given file format) or undef if we hit the end of the file
sub get_alert
{
	my ($fh,$fileformat)= @_;
	# get and parse the next alert in the file, courtesy of snort_alert_parse.pl
	my ($alerttext,$format)= &next_alert($fh,$fileformat);
	return undef unless defined($alerttext); # must have hit end of file
	my $alert= &parse_alert($alerttext,$format);
	$alert->{'type'}= $format;
	# we used to store source file by 'file' => $file_name[$count], but that doesn't seem to be needed anywhere
	return $alert;
}

##############################################################################

# keep records for a new normal alert
sub process_alert
{
	my($alert) = @_;

	# Global stuff
	$earliest_overall = &earliest($alert, $earliest_overall);
	$latest_overall = &latest($alert, $latest_overall);
	$alert_count++;
	
	# Per signature stuff
	$sig_count->{$alert->{'sig'}}++;
	$earliest_sig->{$alert->{'sig'}}
		= &earliest($alert, $earliest_sig->{$alert->{'sig'}});
	$latest_sig->{$alert->{'sig'}}
		= &latest($alert, $latest_sig->{$alert->{'sig'}});

	# Per source stuff
	$src_count->{$alert->{'src'}}++;
	push @{$src_list->{$alert->{'src'}}}, $alert;
	$earliest_src->{$alert->{'src'}}
		= &earliest($alert, $earliest_src->{$alert->{'src'}});
	$latest_src->{$alert->{'src'}}
		= &latest($alert, $latest_src->{$alert->{'src'}});

	# Per dest stuff
	$dest_count->{$alert->{'dest'}}++;
	push @{$dest_list->{$alert->{'dest'}}}, $alert;
	$earliest_dest->{$alert->{'dest'}}
		= &earliest($alert, $earliest_dest->{$alert->{'dest'}});
	$latest_dest->{$alert->{'dest'}}
		= &latest($alert, $latest_dest->{$alert->{'dest'}});
	
	# Per signature-source stuff
	$sig_src_count->{$alert->{'sig'}}{$alert->{'src'}}++;
	
	# Per signature-dest stuff
	$sig_dest_count->{$alert->{'sig'}}{$alert->{'dest'}}++;
	
	# Per source-signature stuff
	$src_sig_count{$alert->{'src'}}{$alert->{'sig'}}++;
	
	# Per dest-signaturestuff
	$dest_sig_count{$alert->{'dest'}}{$alert->{'sig'}}++;
	
	# Src dest stuff
	$src_dest_count->{$alert->{'src'}}{$alert->{'dest'}}++;	 
	$dest_src_count->{$alert->{'dest'}}{$alert->{'src'}}++;	 
		
	# Sig src dest stuff
	$sig_src_dest_count->{$alert->{'sig'}}{$alert->{'src'}}{$alert->{'dest'}}++;
	$sig_dest_src_count->{$alert->{'sig'}}{$alert->{'dest'}}{$alert->{'src'}}++;
}


# keep records for a new anomaly alert
sub process_anom_alert
{
	my($alert) = @_;

	# Global anom stuff
	push(@anom_list,$alert);
	
	# Per source stuff
	$anom_src_count{$alert->{'src'}}++;
	push @{$anom_src_list{$alert->{'src'}}}, $alert;
	$earliest_anom_src->{$alert->{'src'}}
		= &earliest($alert, $earliest_anom_src->{$alert->{'src'}});
	$latest_anom_src->{$alert->{'src'}}
		= &latest($alert, $latest_anom_src->{$alert->{'src'}});

	# Per dest stuff
	$anom_dest_count{$alert->{'dest'}}++;
	push @{$anom_dest_list{$alert->{'dest'}}}, $alert;
	$earliest_anom_dest->{$alert->{'dest'}}
		= &earliest($alert, $earliest_anom_dest->{$alert->{'dest'}});
	$latest_anom_dest->{$alert->{'dest'}}
		= &latest($alert, $latest_anom_dest->{$alert->{'dest'}});
		
	# Src dest stuff
	$anom_src_dest_count{$alert->{'src'}}{$alert->{'dest'}}++;	 
	$anom_dest_src_count{$alert->{'dest'}}{$alert->{'src'}}++;	 
}

##############################################################################

# give each signature a number and find the HTML to use
sub construct_sig_indexes
{
  my($sig,$count);
  
  foreach $sig (keys %$sig_count)
   {
		$sig_index->{$sig} = $count++;
		if($sig =~ /IDS(\d+)/)
		 {
		 	my $num= $1;
		 	$num =~ s/^0+//;
			# ARACHNIDS signature - can make a nice URL for these.
			$sig_url->{$sig} = "http:\/\/whitehats.com\/IDS\/$num";
			$sig_entry->{$sig} = "<a href=\"$sig_url->{$sig}\"".&target('siginfo').">$sig</a>";
		 }
		else
		 { 
			$sig_entry->{$sig} = $sig;
		 }
    }
}

##############################################################################

sub output_main_sig_page
{
	my($sig);
	my $page_title = "Snortsnarf: Snort signatures in $file_name[0] et al";
  
	open(PAGE,">$output_dir$dirsep$index_page") || 
								die("Couldn't open sig page $index_page\n");
	select(PAGE);
	&print_page_head($page_title);
  
	print "<p>$alert_count alerts processed.</p>\n";
	print "<p>Files included:<ul>\n<li>\n";
	print join("<li>\n",@file_name);
	print "\n</ul>\n";
	print "Earliest alert at ".&pretty_time($earliest_overall)."<br>\n";
	print "Latest alert at ".&pretty_time($latest_overall)."</p>\n";
	if (@anom_list) {
		print "<p>The ".+@anom_list." reports from the <A HREF=\"http:\/\/www.silicondefense.com/spice/\">Spade anomaly sensor</A> are in a separte section: <A HREF=\"anomrep/\">visit it</a></p>\n";
	}

	print "<TABLE BORDER CELLPADDING = 5>\n";
	print "<TR><TD>Signature (click for definition)</TD><TD>\# Alerts</TD>".
				"<TD>\# Sources</TD><TD>\# Destinations</TD><TD>Detail link</TD></TR>\n";

	foreach $sig (sort sort_by_sig_count keys %$sig_count)
	{
		print "<TR><TD>$sig_entry->{$sig}</TD><TD>$sig_count->{$sig}</TD><TD>".
				scalar(keys %{$sig_src_count->{$sig}})."</TD><TD>".
				scalar(keys %{$sig_dest_count->{$sig}}).
				"</TD><TD><a href=\"sig$sig_index->{$sig}.$html\">Summary</a></TD></TR>\n";
	}
	print "</TABLE>\n\n";
	if ($db_file ne '') {
		print "<FORM ACTION=\"$cgi_dir/view_annotations.pl\">View/add annotations of type <SELECT NAME=\"type\"><INPUT TYPE=hidden NAME=\"file\" VALUE=\"$db_file\">";
		foreach ('IP','network','snort message') {
			print "<OPTION VALUE=\"$_\"> $_";
		}
		print "</SELECT> for key : <INPUT NAME=\"key\" SIZE=15><INPUT TYPE=\"submit\" VALUE=\"View\"></FORM>";
	}
  
	&print_page_foot();
	close(PAGE);
}

sub sort_by_sig_count { $sig_count->{$a} <=> $sig_count->{$b};}

##############################################################################

# produce the anomaly section, if there are any anomaly reports
sub output_anom_sect {
	return unless @anom_list;
	# clear the anomaly output directory if it exists
	if (-e $anomoutdir) {
	  	if (opendir(D,$anomoutdir)) {
	  		my $file;
	  		while ($file=readdir(D)) {
				next unless $file =~ /^((src|dest)\d+\.\d+\.\d+\.\d+.*|allbyscore|allbytime|srcips|destips|index)\.$html/;
				# looks like a file we created, so delete it
				$fullpath= "$anomoutdir$dirsep$file";
				unlink($fullpath) || warn "could not clear $fullpath in output directory";
			}
			closedir(D);
	  	} else {
			warn "could not open output directory $anomoutdir to clear it";
	  	}
	  }
	unless (-e $anomoutdir && -d $anomoutdir) {
		mkdir($anomoutdir,0755) || die "could not make directory $anomoutdir for the anomaly report section";
	}

	&output_all_anom_page('allbyscore',"score",\&sort_by_score);
	&output_all_anom_page('allbytime',"time",\&sort_by_time);
	&output_anom_src_page();
	&output_anom_dest_page();
	&output_anom_per_source();
	&output_anom_per_dest();
	
	my $page_title = "Snortsnarf: Main Spade report page for $file_name[0] et al";
  
	open(PAGE,">$anomoutdir$dirsep$index_page") || 
								die("Couldn't open main anom page $index_page\n");
	select(PAGE);
	&print_page_head($page_title);
  	&output_anom_header();
	&print_page_foot();
	close(PAGE);

}

sub sort_by_score 
{ 
 $b->{'anomscore'} <=> $a->{'anomscore'};
}


##############################################################################

# output the page for each signature
sub output_per_sig
{
  my($sig,$sig_file,$src,$dest,$early,$late);
  my $page_title;
  
  foreach $sig (keys %$sig_count)
   {  
	# Sort out the file
	$global_sig = $sig;	 # ouch - need to communicate with sort_by_sig_src_count
	$sig_file = "sig$sig_index->{$sig}.$html";

	open(PAGE,">$output_dir$dirsep$sig_file") || 
										die("Couldn't open sig output file $sig_file\n");
	select(PAGE);

	# Print page head stuff
	$page_title = "Summary of alerts in $file_name[0] et al for signature:";
	&print_page_head($page_title);
	print "<h3>$sig_entry->{$sig}</h3>";
	print "<p>$sig_count->{$sig} alerts on this signature.</p>\n";
	print "<p>Looking in files:<ul>\n<li>\n";
	print join("<li>\n",@file_name);
	print "\n</ul>\n";
	print "Earliest such alert at ".&pretty_time($earliest_sig->{$sig})."<br>\n";
	print "Latest such alert at ".&pretty_time($latest_sig->{$sig})."</p>\n";

	# print page head table stuff
	print "<table border cellpadding = 3>\n";
  	print "<tr><td>$sig</td>\n";
  	print "<td><A HREF=#srcsect>",scalar(keys %{$sig_src_count->{$sig}})," sources</A></td>\n";
  	print "<td><A HREF=#destsect>",scalar(keys %{$sig_dest_count->{$sig}})," destinations</A></td></tr>\n";
	if ($db_file ne '') { #link to annotations
		print "<tr><td colspan=3 align=center><A HREF=\"",&view_ann_url('snort message',$sig),"\">View/add annotations for this signature</A></td></tr>\n";
	}
	if (defined($rules_file) && $sig !~ /^(UDP|TCP|ICMP)\s+scan$/) { # show rule file entries for signature
		my(@rules_html)= &get_rules_html_for_sig($sig,$rules_file);
		if (@rules_html) {
			print "<tr bgcolor=\"#D5E2CE\"><td colspan=3 align=center>Rules with message \"$sig\":</td></tr>\n";
			foreach (@rules_html) {
				print "<tr bgcolor=\"#D5E2CE\"><td colspan=3 align=left>$_</td></tr>\n";
			}
		}
	}
	print "</table>";

	# Print the sources section
	print "<hr><h3><A NAME=srcsect>Sources triggering this attack signature</A></h3>\n";
	print "<TABLE BORDER CELLPADDING = 5>\n";
	print "<tr><td>Source</td><td>\# Alerts (sig)</td>".
				"<td>\# Alerts (total)</td><td>\# Dsts (sig)</td>".
				"<td>\# Dsts (total))</td></tr>\n";
	foreach $src (sort sort_by_sig_src_count keys %{$sig_src_count->{$sig}})
		 {
		  print "<tr><td><a href=\"src$src.$html\">$src</a></td>".
				"<td>$sig_src_count->{$sig}{$src}</td>".
				"<td>$src_count->{$src}</td><td>".
				scalar(keys %{$sig_src_dest_count->{$sig}{$src}})."</td><td>".
				scalar(keys %{$src_dest_count->{$src}})."</td></tr>\n";
		 }
	print "</TABLE>\n";
		
	# Print the destinations section
		print "<hr><h3><A NAME=destsect>Destinations receiving this attack signature</A></h3>\n";
		print "<TABLE BORDER CELLPADDING = 5>\n";
		print "<tr><td>Destinations</td><td>\# Alerts (sig)</td>".
				"<td>\# Alerts (total)</td><td>\# Srcs (sig)</td>".
				"<td>\# Srcs (total))</td></tr>\n";
		foreach $dest (sort sort_by_sig_dest_count
											keys %{$sig_dest_count->{$sig}})
		 {
		  print "<tr><td><a href=\"dest$dest.$html\">$dest</a></td>".
				"<td>$sig_dest_count->{$sig}{$dest}</td>".
				"<td>$dest_count->{$dest}</td><td>".
				scalar(keys %{$sig_dest_src_count->{$sig}{$dest}})."</td><td>".
				scalar(keys %{$dest_src_count->{$dest}})."</td></tr>\n";
		 }
	print "</TABLE>\n"; 
	&print_page_foot();
	close(PAGE);
   }
}

sub sort_by_sig_src_count 
{ 
 $sig_src_count->{$global_sig}{$b} <=> 
										$sig_src_count->{$global_sig}{$a};
}

sub sort_by_sig_dest_count 
{ 
 $sig_dest_count->{$global_sig}{$b} <=> 
										$sig_dest_count->{$global_sig}{$a};
}

##############################################################################

# make a page for each source IP
sub output_per_source
{
   my($src,$src_file);
   
   foreach $src (keys %$src_count)
	{  
		&output_per_host($output_dir,$src,'src','general',"from $src in $file_name[0] et al",$src_list->{$src});
	}
}

##############################################################################

# make a page for each destination IP
sub output_per_dest
{
   my($dest,$dest_file);
   
   foreach $dest (keys %$dest_count)
	{  
		&output_per_host($output_dir,$dest,'dest','general',"going to $dest in $file_name[0] et al",$dest_list->{$dest});
	}
}

##############################################################################

# make page(s) listing the alerts for an IP address
sub output_per_host {
	my($outdir,$ip,$end,$pagetype,$al_descr,$alertsref)= @_;
	# this is a very general function.  The args are:
	# + the directory to put the file(s) in
	# + the IP address the page is about
	# + the end this page is about ('src' or 'dest')
	# + the type of page to produce ('general' or 'anom')
	# + the description of the alerts
	# + a reference to the a list of alerts to put on the page
	@alerts= sort sort_by_time @{$alertsref};
	my $num_alerts= @alerts;
	my $ip_file = "$end$ip.$html";
	open(PAGE,">$outdir$dirsep$ip_file") 
		|| die("Couldn't open $ip output file $ip_file\n");
	my $prevsel= select(PAGE);
	
	$al_descr= ($pagetype eq 'general'?'':'anomaly ')."alerts $al_descr";
	if ($split_threshold == 0 || $num_alerts <=  $split_threshold) {
		# start the page
		&print_page_head("All $num_alerts $al_descr");
		&output_per_host_header($ip,$end,$pagetype,$alerts[0],$alerts[$#alerts]);
		&output_alert_table(\@alerts);
	} else {  # need to split
		# make the page containing the overview the alerts
		&print_page_head("Overview of $num_alerts $al_descr");
		&output_per_host_header($ip,$end,$pagetype,$alerts[0],$alerts[$#alerts]);

		my $all_file = "$end$ip-all.$html";

		print "<hr>This listing contains $num_alerts alerts.  You can:\n";
		print "<UL><LI><A HREF=\"$all_file\">view the whole listing</A>\n";
		print "<LI>view a range of alerts <A HREF=#rangelist>(see table below)</A></UL>\n";

		print "<hr><A NAME=rangelist>Alert ranges (sorted by time):</A>\n";
		print "<table border cellpadding = 3><TR align=center><B><TD>alert #'s</TD><TD>first time</TD><TD>last time</TD></B></TR>\n";
		my($first,$last);
		foreach ($first=1; $first < $num_alerts; $first+=$split_threshold) { # segment the alerts
			$last= $first+$split_threshold-1;
			$last= $last < $num_alerts ? $last : $num_alerts;
			my(@alertsub)= @alerts[($first-1)..($last-1)];
			my $early= $alertsub[0];
			my $late= $alertsub[$#alertsub];
			my $range_file = "$end$ip-$first.$html";
			# print this to the "all" page
			print "<tr><td><A HREF=\"$range_file\">$first to $last</A></td><td>",&pretty_time($early),"</td><td>",&pretty_time($late),"</td></tr>\n";

			# create page for the range
			open(RANGE,">$outdir$dirsep$range_file") 
				|| die("Couldn't open $ip output file $range_file\n");
			my $prevsel= select(RANGE);
			&print_page_head("$first to $last of $num_alerts $al_descr");
			&output_per_host_header($ip,$end,$pagetype,$early,$late);
			print "<hr>";
			my $nav= "Go to: ".
				($first == 1?'':"<A HREF=\"$end$ip-".($first-$split_threshold).".$html\">previous range</A>, ").
				(($first+$split_threshold >= $num_alerts)?'':"<A HREF=\"$end$ip-".($first+$split_threshold).".$html\">next range</A>, ").
				" <A HREF=\"$all_file\">all alerts</A>, <A HREF=\"$ip_file\">overview page</A>";
			print $nav;
			&output_alert_table(\@alertsub);
			print $nav;
			&print_page_foot();
			close(RANGE); select($prevsel);
		}		
		print "</table>\n";
		
		# now make the page for all the alerts
		open(ALL,">$outdir$dirsep$all_file")
			|| die("Couldn't open $ip output file $all_file\n");
		my $prevsel= select(ALL);
		&print_page_head("All $num_alerts $al_descr");
		&output_per_host_header($ip,$end,$pagetype,$alerts[0],$alerts[$#alerts]);
		my $nav= "Go to: <A HREF=\"$ip_file\">overview page</A>";
		print $nav;
		&output_alert_table(\@alerts);
		print $nav;
		&print_page_foot();
		close(ALL); select($prevsel);
	}
	
	# finish the page we started with
	&print_page_foot();
	close(PAGE); select($prevsel);
}

##############################################################################

# make the header section at the top of a host page
sub output_per_host_header
{
	my ($ip,$end,$pagetype,$early,$late)=@_;
	# the args are:
	# + the IP address the header is about
	# + the end this page is about ('src' or 'dest')
	# + the type of page to produce ('general' or 'anom')
	# + the earliest alert for the page
	# + the latest alert
	print "<p>Looking in files:<ul>\n<li>\n";
	print join("<li>\n",@file_name);
	print "\n</ul>\n";		
	print "Earliest: ".&pretty_time($early)."<br>\n";
	print "Latest: ".&pretty_time($late)."</p>\n";
	
	&output_anom_pagelist() if $pagetype eq 'anom';
	&output_host_siglist($ip,$end) if $pagetype eq 'general';
	
	# start the table and get the width
	print "<table border cellpadding = 3>\n";
	my $cols= &print_ip_lookup($ip);

	# find the paths to the general and anomaly directories
	my($genpath,$anompath);
	if ($pagetype eq 'general') {
		$genpath='';
		$anompath='anomrep/';
	} else  {
		$genpath='../';
		$anompath='';
	}
	
	my(@seealso)= (); # make a list of "see also"s that we want
	if (!(($pagetype eq 'general') && ($end eq 'src')) && exists $src_count->{$ip}) {
		push(@seealso,"<A HREF=\"$genpath"."src$ip.$html\">an alert source</A>");
	}
	if (!(($pagetype eq 'general') && ($end eq 'dest')) && exists $dest_count->{$ip}) {
		push(@seealso,"<A HREF=\"$genpath"."dest$ip.$html\">an alert destination</A>");
	}
	if (!(($pagetype eq 'anom') && ($end eq 'src')) && exists $anom_src_count{$ip}) {
		push(@seealso,"<A HREF=\"$anompath"."src$ip.$html\">an anomaly source</A>");
	}
	if (!(($pagetype eq 'anom') && ($end eq 'dest')) && exists $anom_dest_count{$ip}) {
		push(@seealso,"<A HREF=\"$anompath"."dest$ip.$html\">an anomaly destination</A>");
	}
	if (@seealso == 1) { # print this with one row
		print "<tr bgcolor=\"#E0CDD0\"><td align=center colspan=$cols>See also $ip as $seealso[0]</td></tr>\n";
	} elsif (@seealso) { # make a common first 2 columns and distinct last columns
		my $colwidth=$cols-2;
		print "<tr bgcolor=\"#E0CDD0\"><td rowspan=".(0+@seealso)." align=center colspan=2>See also $ip as:</td>";
		print "<td colspan=$colwidth align=left>$seealso[0]</td></tr>\n";
		foreach (@seealso[1..$#seealso]) {
			print "<tr bgcolor=\"#E0CDD0\"><td colspan=$colwidth align=left>$_</td></tr>\n";
		}
	}
	
	if ($db_file ne '') {
		# generate links to annotations
		print "<tr><td colspan=$cols align=center><A HREF=\"",&view_ann_url('IP',$ip),"\">View/add annotations for this IP address</A></td></tr>\n";
		$ip =~ /^(\d+\.\d+\.\d+)/;
		my $netkey= "$1.0/24";
		print "<tr><td colspan=$cols align=center><A HREF=\"",&view_ann_url('network',$netkey),"\">View/add annotations for $netkey</A></td></tr>\n";
		$ip =~ /^(\d+\.\d+)/;
		$netkey= "$1.0.0/16";
		print "<tr><td colspan=$cols align=center><A HREF=\"",&view_ann_url('network',$netkey),"\">View/add annotations for $netkey</A></td></tr>\n";
	}
	
	if (defined($nmap_url) && (!defined($nmap_dir) || -e "$nmap_dir/$ip.html")) {
		# generate links to the corresponding nmap2html page
		print "<tr><td colspan=$cols align=center><A HREF=\"$nmap_url$ip.html\">View nmap log page for $ip".(defined($nmap_dir)?'':' (if any)')."</A></td></tr>\n";
	}
	
	if (defined($sisr_config)) {
		# make links to SISR
		my $encconfig= &url_encode($sisr_config);
		# we want to "create" links if there are some anom entries and some normal ones for this IP and end 
		my $showcreate2= ($pagetype eq 'anom') && ($end eq 'src') && exists $src_count->{$ip};
		$showcreate2 ||= ($pagetype eq 'anom') && ($end eq 'dest') && exists $dest_count->{$ip};
		$showcreate2 ||= ($pagetype eq 'general') && ($end eq 'src') && exists $anom_src_count{$ip};
		$showcreate2 ||= ($pagetype eq 'general') && ($end eq 'dest') && exists $anom_dest_count{$ip};
		my $rows= 3+ ($showcreate2 ? 1 : 0);
		my $colwidth=$cols-1;
		print "<tr bgcolor=\"#D5E2CE\"><td rowspan=$rows align=center>Incident handling</td>";
		my $endip= ($end eq 'src') ? "from $ip" : "to $ip";
		my $linktext= "Add some the ".(@anom_list?(($pagetype eq 'general')?'standard ':'anomaly '):'')." alerts $endip to a labeled set";
		print "<td colspan=$colwidth align=center><A HREF=\"$cgi_dir/sel_to_add.pl?".join('&',"configfile=$encconfig","end=$end","ip=$ip",'include='.($pagetype eq 'general'?'g':'a'),'logs='.&url_encode(join(',',@file_fullpath))).'"',&target('sisrwin'),">$linktext</A></td></tr>\n";
		if ($showcreate2) {
			$linktext= "Add some of both type alerts $endip to a labeled set";
			print "<tr bgcolor=\"#D5E2CE\"><td colspan=$colwidth align=center><A HREF=\"$cgi_dir/sel_to_add.pl?".join('&',"configfile=$encconfig","end=$end","ip=$ip",'include=ga','logs='.&url_encode(join(',',@file_fullpath))).'"',&target('sisrwin'),">$linktext</A></td></tr>\n";
		}
		print "<tr bgcolor=\"#D5E2CE\"><td colspan=$colwidth align=center><A HREF=\"$cgi_dir/lsetlist.pl?configfile=$encconfig\"",&target('sisrwin'),">List stored sets</td></tr>\n";
		print "<tr bgcolor=\"#D5E2CE\"><td colspan=$colwidth align=center><A HREF=\"$cgi_dir/inclist.pl?configfile=$encconfig\"",&target('sisrwin'),">List stored incidents</td></tr>\n";
	}
	print "</table>\n";
}


sub output_host_siglist {
	my($ip,$end)= @_;
	my($sighash)= $end eq 'src' ? $src_sig_count{$ip} : $dest_sig_count{$ip};
	my @sigs= keys %{$sighash};
	print "<p>",0+@sigs," different signatures are present for <EM>$ip</EM> as a ",($end eq 'src' ? 'source' : 'destination'),"\n<UL>";
	foreach (sort {$sighash->{$a} <=> $sighash->{$b}} @sigs) {
		print "<LI>$sighash->{$_} instances of <a href=\"sig$sig_index->{$_}.$html\"><EM>$_</EM></A></LI>\n";
	}
	print "</UL></P>";
}

# return a URL to view annotations for and type and key
sub view_ann_url {
	my($type,$key)= @_;
	return "$cgi_dir/view_annotations.pl\?".join('&','file='.&url_encode($db_file),'type='.&url_encode($type),'key='.&url_encode($key));
}


##############################################################################
# make a page listing all anomaly alerts
sub output_all_anom_page {
	my($file,$sortbytext,$sortfn)= @_;
	# args are:
	# + the file name to give the page
	# + the text for "sorted by..."
	# + a function to use to sort the alerts
	my @alerts= sort $sortfn @anom_list;
	my $num_alerts= @alerts;
	open(PAGE,">$anomoutdir$dirsep$file.$html") 
		|| die("Couldn't open output file $file.$html\n");
	my $prevsel= select(PAGE);
		
	&print_page_head("All ".scalar(@anom_list)." Anomaly reports, sorted by $sortbytext");
	&output_anom_header();
	&output_alert_table(\@alerts);
	
	&print_page_foot();
	close(PAGE); select($prevsel);
}

# print out the anomaly specific header
sub output_anom_header {
	my $totcount= @anom_list;
	print "<p>$totcount anomaly reports found among the files:<ul>\n<li>\n";
	print join("\n<li>",@file_name);
	print "\n</ul>\n";
	&output_anom_pagelist();
}

sub output_anom_pagelist {
	my $srccount= keys %anom_src_list;
	my $destcount= keys %anom_dest_list;
	print <<">>";
<P>There are 4 top level pages for alerts produced by the <A HREF=\"http:\/\/www.silicondefense.com/spice/\">Spade anomaly sensor</A>:
<UL>
	<LI><A HREF="allbyscore.$html">All alerts sorted by score
	<LI><A HREF="allbytime.$html">All alerts sorted by time
	<LI><A HREF="srcips.$html">A list of the $srccount source IP addresses in the alerts
	<LI><A HREF="destips.$html">A list of the $destcount destination IP addresses in the alerts
</UL>
<P><A HREF="../">See also the main signature page</A>
>>
}

##############################################################################
# display the pages with anomaly source IPS
sub output_anom_src_page
{
	my $file = "srcips.$html";

	open(PAGE,">$anomoutdir$dirsep$file") || 
										die("Couldn't open output file $file\n");
	select(PAGE);

	# Print page head stuff
	my $page_title = "Source IPs of anomaly reports in $file_name[0] et al";
	&print_page_head($page_title);
	&output_anom_header();
	my $ipcount= keys %anom_src_list;
	print "<p>$ipcount distinct source IPs are present in the anomaly reports.</p>\n";

	# Print the sources section
	print "<hr><h3><A NAME=srcsect>Source IPs in anomaly reports</A></h3>\n";
	print "<TABLE BORDER CELLPADDING = 5>\n";
	print "<tr><td>Source</td><td>\# Alerts</td>".
				"<td>\# Anom Dsts</td>".
				"<td>\# Other Alerts </td></tr>\n";
	my $src;
	foreach $src (sort sort_by_anom_src_count keys %anom_src_list) {
		print "<tr><td><a href=\"src$src.$html\">$src</a></td>".
			"<td>$anom_src_count{$src}</td><td>".
			scalar(keys %{$anom_src_dest_count{$src}})."</td>".
			"<td>".(0+($src_count->{$src}))."</td></tr>\n";
	}
	print "</TABLE>\n";
		
	&print_page_foot();
	close(PAGE);
}

sub sort_by_anom_src_count 
{ 
 $anom_src_count{$b} <=> $anom_src_count{$a};
}

# display the pages with anomaly destination IPS
sub output_anom_dest_page
{
	my $file = "destips.$html";

	open(PAGE,">$anomoutdir$dirsep$file") || 
										die("Couldn't open output file $file\n");
	select(PAGE);

	# Print page head stuff
	my $page_title = "Destination IPs of anomaly reports in $file_name[0] et al";
	&print_page_head($page_title);
	&output_anom_header();
	my $ipcount= keys %anom_dest_list;
	print "<p>$ipcount distinct destination IPs are present in the anomaly reports.</p>\n";

	# Print the dests section
	print "<hr><h3><A NAME=srcsect>Destination IPs in anomaly reports</A></h3>\n";
	print "<TABLE BORDER CELLPADDING = 5>\n";
	print "<tr><td>Destination</td><td>\# Alerts</td>".
				"<td>\# Anom Srcs</td>".
				"<td>\# Other Alerts </td></tr>\n";
	my $dest;
	foreach $dest (sort sort_by_anom_dest_count keys %anom_dest_list) {
		print "<tr><td><a href=\"dest$dest.$html\">$dest</a></td>".
			"<td>$anom_dest_count{$dest}</td><td>".
			scalar(keys %{$anom_dest_src_count{$dest}})."</td>".
			"<td>".(0+($dest_count->{$dest}))."</td></tr>\n";
	}
	print "</TABLE>\n";
		
	&print_page_foot();
	close(PAGE);
}

sub sort_by_anom_dest_count 
{ 
 $anom_dest_count{$b} <=> $anom_dest_count{$a};
}

##############################################################################

# make a page for each anomaly source IP
sub output_anom_per_source
{
	my($src);
   
	foreach $src (keys %anom_src_list) {  
		&output_per_host($anomoutdir,$src,'src','anom',"from $src in $file_name[0] et al",$anom_src_list{$src});
	}
}

# make a page for each anomaly destination IP
sub output_anom_per_dest
{
	my($dest);
   
	foreach $dest (keys %anom_dest_list) {  
		&output_per_host($anomoutdir,$dest,'dest','anom',"to $dest in $file_name[0] et al",$anom_dest_list{$dest});
	}
}


##############################################################################

# output a table of alerts
sub output_alert_table {
	my($alertsref)= @_;
	print "<HR>";
	print "<table border cellpadding = 3>\n";
	for (my $i=0; $i <= $#{$alertsref}; $i++) {
		&output_table_entry($alertsref->[$i]);
	 }
	print "</table>\n";
}

# output a table entry for a given alert
sub output_table_entry { 
	my($alert)= @_;
	my  $tdopts= $color_opt eq 'rotate'?" bgcolor=".&alert_color($alert):'';
	print "<tr><td$tdopts>".&alert_as_html($alert)."</td></tr>\n";
}

##############################################################################

# find a color to use as the background for an alert
sub alert_color { 
	my($alert)= @_;
	return $color[$::old_color] if (defined($::old_alert) && (($::old_alert->{'sig'} eq $alert->{'sig'}) || defined($alert->{'anomscore'})) && ($::old_alert->{'src'} eq $alert->{'src'}) && ($::old_alert->{'dest'} eq $alert->{'dest'}));
	$::old_color= ++$::old_color % +@color;
	$::old_alert= $alert;
	return $color[$::old_color];
}

##############################################################################

# make an alert into HTML
sub alert_as_html { 
	my($alert)= @_;
	$append= '';
	$text= $alert->{'text'};
    unless (defined($alert->{'anomscore'})) {
	    my $sig= $alert->{'sig'};
	    my $sigre= $sig;
	    $sigre =~ s/([^\w ])/\\$1/g;
	    $text =~ s/(\[\*\*\]\s*)($sigre)(\s*\[\*\*\])/$1<a href=\"sig$sig_index->{$sig}.$html\">$2<\/A>$3/;
	}
	$text =~ s/(\d+\.\d+\.\d+\.\d+)(.*)->/<A HREF="src$1.html">$1<\/A>$2->/;
	$text =~ s/->(\s*)(\d+\.\d+\.\d+\.\d+)/->$1<A HREF="dest$2.html">$2<\/A>/;
	$text =~ s/[\n\r]+/<br>/g;
    if ($log_base ne '' && ($alert->{'type'} =~ /alert$/) && defined($alert->{'proto'})) {
       $append.= " <A HREF=\"".&get_alert_logpage($alert)."\">[Snort log]</A>\n";
    } 
	return "<code>$text</code>$append";
}

##############################################################################

# text to add to use a given locatation as a target link or '' if -onewindow was given
sub target {
	return '' if $notarget_option;
	return " target=$_[0]";
}

##############################################################################

# given a snort message, dig through the given rules file and generate HTML
# for each line than has that signature.  A list is produced and included
# files are followed
sub get_rules_html_for_sig {
	my ($sig,$rule_file)= @_;
	my(@rules_html)=();
	my $fh= $fhnonce++;
	# file names assumed to be absolute or relative to the current directory
	# but if -rulesdir was given, always use that as the dirctory (with the
	# file name appended to make the file location)
	$rule_file =~ s/\s+$//;
	my($file)= $rule_file =~ /([^\/]+)$/; 
	if (defined($rules_dir)) {
		$rule_file= "$rules_dir/$file";
	}
	unless (open($fh,"<$rule_file")) {
		warn "could not open $rule_file to read rules from -- skipping\n";
		return;
	}
	while (<$fh>) {
		chomp;
		next if /^(\#|\s*$)/;
		if (s/^\s*include\s+//) {
			s/\s+$//;
			push(@rules_html,&get_rules_html_for_sig($sig,$_));
		} elsif (/\(.*msg\s*:\s*\"([^\"]*)\"/) {
			my $mess= $1;
			$mess =~ s/\\(.)/$1/g;
			$mess =~ s/^\s+//;
			$mess =~ s/\s+$//;
			if ($mess eq $sig) { # message in rule is same as one looked for (modulo whitespace)
				my $html= "<SMALL>$_</SMALL> (from <EM>$file</EM>)";
				push(@rules_html,$html);
			}
		}
	}
	close $fh;
	return @rules_html;
}

##############################################################################

# return the page in the full snort logs that corresponds to a given alert
sub get_alert_logpage 
{
	my($alert)=@_;
	my $sport=$alert->{'sport'};
	my $dport=$alert->{'dport'};
	my $src=$alert->{'src'};
	my $dest=$alert->{'dest'};
	my $proto=$alert->{'proto'};
	my ($ip,$port1,$port2);

	my $srcishome= &in_homenet($src);
	my $destishome= &in_homenet($dest);

	if ($destishome && !$srcishome) {
		$ip= $src; $port1= $sport; $port2= $dport;
	} elsif ($srcishome && !$destishome) {
		$ip= $dest; $port1= $dport; $port2= $sport;
	} elsif ($sport >= $dport) {
		$ip= $src; $port1= $sport; $port2= $dport;
	} else {
		$ip= $dest; $port1= $dport; $port2= $sport;
	}
   
	if ($proto eq 'ICMP') {
	   my $ICMP_type= $alert->{'ICMP_type'};
		print STDOUT "Warning: \"$ICMP_type\" text not found in \%ICMP_text_to_filename table\n" unless defined($ICMP_text_to_filename{$ICMP_type});
		return $log_base.$ip.'/'.$ICMP_text_to_filename{$ICMP_type};
	} else {
		my $prototext= ($proto eq 'UDP')?'UDP':'TCP';
		return $log_base.$ip."/".$prototext.":".$port1."-".$port2;
	} 
}

##############################################################################

# print the line of a table to look up an IP address in whois databases;
# returns the number of columns used
sub print_ip_lookup
{
  my($ip) = @_;
 
  my $host = gethostbyaddr(inet_aton($ip), AF_INET) if defined $dns_option;
  
  print "<tr><td>$ip</td>\n";
  print "<td>($host)</td>\n" if defined $host;
  print "<td>Lookup at:</td>\n";
  print "<td><a href=\"http://www.arin.net/cgi-bin/whois.pl".
				"?queryinput=$ip&B1=Submit+Query\"",&target('lookup'),">ARIN</a></td>\n";
  print "<td><a href=\"http://www.ripe.net/cgi-bin/whois".
				"?query=$ip&.=Submit+Query\"",&target('lookup'),">RIPE</a></td>\n";
  print "<td><a href=\"http://www.apnic.net/apnic-bin/whois.pl".
				"?search=$ip\"",&target('lookup'),">APNIC</a></td>\n";
  print "<td><a href=\"http://www.geektools.com/cgi-bin/proxy.cgi".
  				"?query=$ip&targetnic=auto\"",&target('lookup'),">Geektools</a></td>\n";
  				# thanks to Dr. Paul Mitchell for this add
  print "</tr>\n";
  return 6 + (defined $host ? 1: 0); # (number of cols used)

#				 alias jpnic  "/usr/ucb/whois -h whois.nic.ad.jp"
#				 alias aunic  "/usr/ucb/whois -h whois.aunic.net"
#				 alias milnic "/usr/ucb/whois -h whois.nic.mil"
#				 alias govnic "/usr/ucb/whois -h whois.nic.gov"
#				 alias krnic  "/usr/ucb/whois -h whois.krnic.net"
}

##############################################################################

# prints out a standard SnortSnarf HTML header
sub print_page_head
{
  my($page_title) = @_;
  
  print "<html>\n<head>\n<title>$page_title</title>\n";
  print "</head>\n<body BGCOLOR=\"$bgcol\">\n";
  print "<h2>$page_title</h2>\n\n";
}

##############################################################################

# prints out a standard SnortSnarf HTML footer
sub print_page_foot
{
 print "<hr>\n";
 print "<i>Generated by $script $version ($author)</i><p>\n";
 print "See also the <a href=\"http://www.snort.org/\">".
						"Snort Page</a> by Marty Roesch<p>";
 print "Page generated at ".localtime(time())."<p>\n";
 print "</body></html>\n";
}

##############################################################################

# sort $a and $b (parsed alerts) by time, returning a <=>/cmp type result
sub sort_by_time
{  
  my(@pieces1) = ($a->{'month'},$a->{'date'},split(':',$a->{'time'}));
  my(@pieces2) = ($b->{'month'},$b->{'date'},split(':',$b->{'time'}));
  
  foreach (0..$#pieces1)
   {
	return -1 if $pieces1[$_] < $pieces2[$_];
	return 1 if $pieces1[$_] > $pieces2[$_];
   }
 return 0;
}

##############################################################################

# find the earlier of two alerts and return it
sub earliest
{
 my($alert1,$alert2) = @_;
 
 return $alert1 unless defined $alert2;
 return $alert2 unless defined $alert1;
 
 my(@pieces1) = ($alert1->{'month'},$alert1->{'date'},split(':',$alert1->{'time'}));
 my(@pieces2) = ($alert2->{'month'},$alert2->{'date'},split(':',$alert2->{'time'}));
 
 foreach (0..$#pieces1)
  {
   return $alert1 if $pieces1[$_] < $pieces2[$_];
   return $alert2 if $pieces1[$_] > $pieces2[$_];
  }
 return $alert1;
}

##############################################################################

# find the later of two alerts and return it
sub latest
{
 my($alert1,$alert2) = @_;
 
 return $alert1 unless defined $alert2;
 return $alert2 unless defined $alert1;
 
 my(@pieces1) = ($alert1->{'month'},$alert1->{'date'},split(':',$alert1->{'time'}));
 my(@pieces2) = ($alert2->{'month'},$alert2->{'date'},split(':',$alert2->{'time'}));
 
 foreach (0..$#pieces1)
  {
   return $alert1 if $pieces1[$_] > $pieces2[$_];
   return $alert2 if $pieces1[$_] < $pieces2[$_];  
  }
 return $alert2;
}

##############################################################################

# return HTML to pretty-print the alert's time
sub pretty_time
{
  my($alert) = @_;
  my(@bits) = split(/\./,$alert->{'time'});
  return "<b>$bits[0]</b>".
						(defined($bits[1])&&($bits[1] ne '')?".$bits[1]":'').
						" <i>on $alert->{'month'}/$alert->{'date'}</i>";
}

##############################################################################

# make a bit mask/array out of a list of 4 byte-size ints, where the first
#  element is the most significant
# really this should adjust for big endian/little endian, but this is fine
#  for our purposes
sub bytenums2bits {
	return ($_[0] << 24) | ($_[1] << 16) | ($_[2] << 8) | $_[3];
}

sub in_homenet {
	return 0 unless defined($homenetaddr);
	my $ipaddr= &bytenums2bits(split('\.',$_[0]));
	return ($ipaddr & $homenetmask) == $homenetaddr;
}

##############################################################################

# only used in debugging, print full details of a parsed alert
sub debug_print_alert {
	foreach $key (keys %{$alert}) {
		print STDOUT "	$key: ",$alert->{$key},"\n";
	}
	print STDOUT "\n";
}
